from .errors import ApiError, RTRError, RTRConnectionError
from ..log import log


class RTRSession:
    def __init__(self, falcon_api, device_id):
        self.falcon = falcon_api
        self.device_id = device_id
        self.session = self._connect()

    def close(self):
        # return self.falcon.client.command('RTR_DeleteSession', session_id=self.id)
        # Below is a workaround for the above not working properly (/cc @jshcodes)

        from falconpy import OAuth2, RealTimeResponse  # pylint: disable=C0415
        from ..config import config  # pylint: disable=C0415
        falcon_auth = OAuth2(
            client_id=config.get('falcon', 'client_id'),
            client_secret=config.get('falcon', 'client_secret')
        )
        falcon_rtr = RealTimeResponse(falcon_auth)
        response = falcon_rtr.delete_session(session_id=self.id)
        if response['status_code'] != 204:
            log.debug('Unable to close the RTR session: reponse was: %s', response)

    def execute_and_wait(self, action, base_command, command_string):
        command = self._execute(action, base_command, command_string)
        response = self._rtr_wait(command[0])

        if response['stderr']:
            raise RTRError(f'RTR Execute device: {self.device_id}, stderr: {response["stderr"]}, (session_id: {response["session_id"]}, task_id: {response["task_id"]})')
        return response

    def get_file(self, filepath):
        command = self.execute_and_wait('RTR_ExecuteActiveResponderCommand', 'get', 'get ' + filepath)
        files = self._list_files()
        for f in files:
            if f['cloud_request_id'] == command['task_id']:
                return self._fetch_file(f['sha256'], filepath)

        log.error('RTR File Not Found. Searching for cloud_request_id=%s, but found only %s', command["task_id"], files)
        raise RTRError(f'RTR File Not Found: device: {self.device_id}, file: {filepath} (session_id: {command["session_id"]}, task_id: {command["session_id"]})')

    @property
    def id(self):
        return self.session['session_id']

    def _list_files(self):
        return self.falcon._resources(  # pylint: disable=W0212
            'RTR_ListFiles',
            parameters={
                'session_id': self.id
            }
        )

    def _fetch_file(self, sha256, filepath):
        response = self.falcon.client.command(
            'RTR_GetExtractedFileContents',
            parameters={
                'session_id': self.id,
                'sha256': sha256,
                'filepath': filepath,
            }
        )
        if not isinstance(response, (bytes, bytearray)):
            raise RTRError(f"Could not fetch RTR file from Falcon: {response['body']}")
        return response

    def _connect(self):
        response = None
        try:
            response = self.falcon.init_rtr_session(self.device_id)
        except ApiError as e:
            raise RTRConnectionError(f"{e}") from e

        if len(response) != 1:
            raise RTRError(f'Unexpected response from RTR Init: {response}')
        return response[0]

    def _rtr_wait(self, command):
        check = self.falcon.check_rtr_command_status(command['cloud_request_id'], 0)[0]
        while not check['complete']:
            check = self.falcon.check_rtr_command_status(command['cloud_request_id'], 0)[0]
        return check

    def _execute(self, action, base_command, command_string):
        return self.falcon.execute_rtr_command(action, self.id, base_command, command_string)



import threading
from ..log import log


class StoppableThread(threading.Thread):
    def __init__(self, *args, stop_event=threading.Event(), **kwargs):
        super().__init__(*args, **kwargs)
        self.stop_event = stop_event

    @property
    def stopped(self):
        return self.stop_event.is_set()

    def stop(self):
        if not self.stopped:
            log.info("Thread requested to stop")
            self.stop_event.set()



import sys
import datetime
import logging
import time
import threading
import requests

from .api import FalconAPI, NoStreamsError
from .models import Event, Stream
from ..util import StoppableThread
from ..log import log
from ..config import config


class StreamManagementThread(threading.Thread):
    """Thread that spins-out sub-threads to manage CrowdStrike Falcon Streaming API"""

    def __init__(self, output_queue, relevant_event_types, *args, **kwargs):
        kwargs['name'] = kwargs.get('name', 'cs_mngmt')
        super().__init__(*args, **kwargs)
        self.output_queue = output_queue
        self.relevant_event_types = relevant_event_types
        self.application_id = config.get('falcon', 'application_id')

    def run(self):
        while True:
            try:
                workers_died_event = self.start_workers()
                while not workers_died_event.is_set():
                    time.sleep(60)
                log.debug("Restarting stream session")
            except Exception:  # pylint: disable=broad-except
                log.exception("Could not restart stream session")
                sys.exit(1)  # TODO implement re-try mechanism

    def start_workers(self):
        stop_event = threading.Event()
        falcon_api = FalconAPI()
        for stream in self.get_streams(falcon_api):
            StreamingThread(stream, self.output_queue, self.relevant_event_types, stop_event=stop_event).start()
            StreamRefreshThread(self.application_id, stream, falcon_api, stop_event=stop_event).start()
        return stop_event

    def get_streams(self, falcon_api):
        retry_count = int(config.get('falcon', 'reconnect_retry_count'))
        for attempt in range(retry_count):
            try:
                return falcon_api.streams(self.application_id)
            except NoStreamsError:
                if attempt == (retry_count - 1):
                    raise
                log.info("Falcon returns no available streams. Re-trying in 10 seconds.")
                time.sleep(10)
        raise NoStreamsError(self.application_id)


class StreamRefreshThread(StoppableThread):
    def __init__(self, application_id, stream: Stream, falcon_api: FalconAPI, *args, **kwargs):
        kwargs['name'] = kwargs.get('name', 'cs_refresh')
        super().__init__(*args, **kwargs)
        self.stream = stream
        self.falcon_api = falcon_api
        self.application_id = application_id

    def run(self):
        try:
            self.sleep()
            while not self.stopped:
                self.refresh_stream_session()
                self.sleep()

        finally:
            self.stop()

    def sleep(self):
        time.sleep(self.stream.refresh_interval * 9 / 10)

    def refresh_stream_session(self):
        self.falcon_api.refresh_streaming_session(self.application_id, self.stream)
        log.debug("Refresh of streaming session succeeded")


class StreamingThread(StoppableThread):
    def __init__(self, stream: Stream, queue, relevant_event_types, *args, **kwargs):
        kwargs['name'] = kwargs.get('name', 'cs_stream')
        super().__init__(*args, **kwargs)
        self.stream = stream
        self.conn = StreamingConnection(self.stream, queue.last_offset(self.stream.feed_id))
        self.queue = queue
        self.relevant_event_types = relevant_event_types
        self.event_count = 0
        self.event_count_types = {}

    def run(self):
        try:
            for event in self.conn.events():
                if event:
                    self.process_event(event)

                if self.stopped:
                    break
        except requests.exceptions.ChunkedEncodingError:
            pass  # ChunkedEncodingError is expected when streaming session closes abruptly
        finally:
            log.warning("Streaming Connection was closed.")
            if not self.stopped:
                self.stop()
            else:
                self.conn.close()

    def process_event(self, event):
        event = Event(event, self.stream.feed_id)
        if log.level <= logging.DEBUG:
            self.log_event(event)

        if (self.relevant_event_types is None or event.event_type in self.relevant_event_types) and not event.irrelevant():
            self.queue.put(event)

    def log_event(self, event):
        self.event_count += 1
        self.event_count_types[event.event_type] = self.event_count_types.get(event.event_type, 0) + 1
        if self.event_count % 200 == 0:
            log.debug("Received %d events from the stream. Type breakdown was: %s", self.event_count, self.event_count_types)


class StreamingConnection():
    def __init__(self, stream: Stream, last_seen_offset=0):
        self.stream = stream
        self.connection = None
        self.last_seen_offset = last_seen_offset

    def open(self):
        headers = {
            'Authorization': 'Token %s' % (self.stream.token),
            'Date': datetime.datetime.utcnow().strftime('%a, %d %b %Y %H:%M:%S +0000'),
            'Connection': 'Keep-Alive'
        }
        log.info("Opening Streaming Connection")
        url = self.stream.url + '&offset={}'.format(self.last_seen_offset + 1 if self.last_seen_offset != 0 else 0)
        self.connection = requests.get(url, headers=headers, stream=True, timeout=60)
        log.info("Established Streaming Connection: %d %s", self.connection.status_code, self.connection.reason)
        self.connection.raise_for_status()
        return self.connection

    def events(self):
        return self.open().iter_lines()

    def close(self):
        if self.connection:
            self.connection.close()
            self.connection = None



from base64 import b64decode, b64encode
from hashlib import sha256
from datetime import datetime
from json import dumps
from hmac import new
from requests import post
from ...log import log
from ...config import config
from ...falcon.errors import RTRConnectionError


def build_signature(workspace_id, primary_key, date, content_length, method, content_type, resource):
    x_headers = 'x-ms-date:' + date
    string_to_hash = method + "\n" + str(content_length) + "\n" + content_type + "\n" + x_headers + "\n" + resource
    bytes_to_hash = bytes(string_to_hash, encoding="utf-8")
    decoded_key = b64decode(primary_key)
    encoded_hash = b64encode(new(decoded_key, bytes_to_hash, digestmod=sha256).digest()).decode()
    authorization = "SharedKey {}:{}".format(workspace_id, encoded_hash)
    return authorization


def post_data(workspace_id, primary_key, body, log_type):
    method = 'POST'
    content_type = 'application/json'
    resource = '/api/logs'
    rfc1123date = datetime.utcnow().strftime('%a, %d %b %Y %H:%M:%S GMT')
    content_length = len(body)
    signature = build_signature(
        workspace_id,
        primary_key,
        rfc1123date,
        content_length,
        method,
        content_type,
        resource)
    uri = 'https://' + workspace_id + '.ods.opinsights.azure.com' + resource + '?api-version=2016-04-01'
    headers = {
        'content-type': content_type,
        'Authorization': signature,
        'Log-Type': log_type,
        'x-ms-date': rfc1123date
    }
    response = post(uri, data=body, headers=headers, timeout=60)
    if (response.status_code < 200 or response.status_code > 299):
        log.error("Failed to send detection to Log Analytics: %s", response.text)


class Submitter():
    AZURE_ARC_KEYS = ['resourceName', 'resourceGroup', 'subscriptionId', 'tenantId', 'vmId']

    def __init__(self, event):
        self.event = event
        self.workspace_id = config.get('azure', 'workspace_id')
        self.primary_key = config.get('azure', 'primary_key')
        self.azure_arc_config = self.autodiscovery()

    def autodiscovery(self):
        if self.event.cloud_provider == 'AZURE' or not config.getboolean('azure', 'arc_autodiscovery'):
            return None

        if self.event.device_details['platform_name'] not in ['Linux', 'Windows']:
            log.info('Skipping Azure Arc Autodiscovery for %s (aid=%s, name=%s)',
                     self.event.device_details['platform_name'],
                     self.event.original_event.sensor_id,
                     self.event.original_event.computer_name
                     )
            return None
        if self.event.device_details['product_type_desc'] == 'Pod':
            log.info('Skipping Azure Arc Autodiscovery for k8s pod (aid=%s, name=%s)',
                     self.event.original_event.sensor_id,
                     self.event.original_event.computer_name
                     )
            return None

        try:
            azure_arc_config = self.event.azure_arc_config()
        except RTRConnectionError as e:
            log.error("Cannot fetch Azure Arc info from host (aid=%s, hostname=%s, last_seen=%s): %s",
                      self.event.original_event.sensor_id,
                      self.event.device_details['hostname'],
                      self.event.device_details['last_seen'],
                      e
                      )
            return None
        except Exception as e:  # pylint: disable=W0703
            log.exception("Cannot fetch Azure Arc info from host (aid=%s, hostname=%s, last_seen=%s): %s",
                          self.event.original_event.sensor_id,
                          self.event.device_details['hostname'],
                          self.event.device_details['last_seen'],
                          e
                          )
            return None

        return {k: v
                for k, v in azure_arc_config.items()
                if k in self.AZURE_ARC_KEYS
                }

    def submit(self):
        log.info("Processing detection: %s", self.event.detect_description)
        post_data(self.workspace_id, self.primary_key, self.log(), 'FalconIntegrationGatewayLogs')

    def log(self):
        json_data = [{
            'ExternalUri': self.event.falcon_link,
            'FalconEventId': self.event.event_id,
            'ComputerName': self.event.original_event.computer_name,
            'Description': self.event.detect_description,
            'Severity': self.event.severity,
            'Title': 'Falcon Alert. Instance {}'.format(self.event.instance_id),
            'ProcessName': self.event.original_event['event']['FileName'],
            'ProcessPath': self.event.original_event['event']['FilePath'],
            'CommandLine': self.event.original_event['event']['CommandLine'],
            'DetectName': self.event.detect_name,
            'AccountId': self.event.cloud_provider_account_id,
            'InstanceId': self.event.instance_id,
            'CloudProvider': self.cloud,
            'ResourceGroup': self.event.device_details.get('zone_group', None)
        }]

        if self.azure_arc_config is not None:
            json_data[0]['arc'] = self.azure_arc_config

        return dumps(json_data)

    @property
    def cloud(self):
        return self.event.cloud_provider if self.event.cloud_provider is not None else 'Unrecognized'


class Runtime():
    RELEVANT_EVENT_TYPES = ['DetectionSummaryEvent']

    def __init__(self):
        log.info("Azure Backend is enabled.")

    def is_relevant(self, falcon_event):
        return True

    def process(self, falcon_event):
        Submitter(falcon_event).submit()


__all__ = ['Runtime']



import re
from functools import lru_cache
from google.cloud.securitycenter import Asset, Finding, SecurityCenterClient
import google.api_core.exceptions
from ...log import log
from . import api


class APIDataError(Exception):
    pass


class AssetNotFound(APIDataError):
    pass


class Cache():
    def __init__(self):
        self._projects = {}
        self._sources = {}
        self._assets = {}
        self._findings = {}

    def asset(self, event):
        asset_id = event.instance_id

        if asset_id not in self._assets:
            project_number = event.cloud_provider_account_id
            assets = self.scc.get_asset(project_number, asset_id)
            if len(assets) == 1:
                self._assets[asset_id] = assets[0].asset
            elif len(assets) == 0:
                raise AssetNotFound("Asset {} not found in GCP Project {}".format(asset_id, project_number))
            else:
                raise APIDataError(
                    "Multiple assets found with ID={} within GCP Project {}".format(asset_id, project_number))
        return self._assets[asset_id]

    def source(self, org_id):
        if org_id not in self._sources:
            self._sources[org_id] = self.scc.get_or_create_fig_source(org_id)
        return self._sources[org_id]

    def organization_parent_of(self, project_id):
        project = self.project(project_id)
        return api.project_get_parent_org(project)

    def project_number_accesible(self, project_number: int) -> bool:
        try:
            return self.project(project_number) is not None
        except google.api_core.exceptions.PermissionDenied:
            return False

    def project(self, project_number: int):
        if project_number not in self._projects:
            self._projects[project_number] = api.project(project_number)
        return self._projects[project_number]

    def submit_finding(self, finding_id, finding: Finding, org_id: str):
        if org_id not in self._findings:
            self._findings[org_id] = {}

        if finding_id in self._findings[org_id]:
            log.debug("Finding %s already exists in GCP SCC", finding_id)
            return None

        finding = self.scc.get_or_create_finding(finding_id, finding, self.source(org_id))
        self._findings[org_id][finding_id] = finding
        return finding

    @property
    @lru_cache
    def scc(self):
        return api.SecurityCommandCenter()


class Submitter():
    def __init__(self, cache, event):
        self.cache = cache
        self.event = event

    def submit(self):
        log.info("Processing detection: %s", self.event.detect_description)
        if not self.cache.project_number_accesible(self.gcp_project_number):
            log.warning(
                "Cannot access GCP project (number=%s) to report malicious behaviour to GCP Security Command Center. Please grant 'roles/securitycenter.admin' role to this service account in every GCP organization that needs to have CrowdStrike detections forwarded to SCC.",
                self.gcp_project_number)
            return

        try:
            finding = self.finding()
        except AssetNotFound:
            log.warning("Corresponding asset not found in GCP Project")
            return

        self.submit_finding(finding)

    def finding(self):
        return Finding(
            name=self.finding_path,
            parent=self.source_path,
            resource_name=self.asset.security_center_properties.resource_name,
            state=Finding.State.ACTIVE,
            external_uri=self.event.falcon_link,
            event_time=self.event.time,
            category=self.event_category,
            severity=self.severity,
            source_properties={
                'FalconEventId': self.event.event_id,
                'ComputerName': self.event.original_event.computer_name,
                'Description': self.event.detect_description,
                'Severity': self.severity,
                'Title': 'Falcon Alert. Instance {}'.format(self.event.instance_id),
                'Category': self.event_category,
                'ProcessInformation': {
                    'ProcessName': self.event.original_event['event']['FileName'],
                    'ProcessPath': self.event.original_event['event']['FilePath'],
                    'CommandLine': self.event.original_event['event']['CommandLine']
                },
            }
        )
        # ART Uncomment these fields to force behavior parity with the prior art. Don't.
        # ART finding.severity=self.original_event['event']['Severity']
        # ART finding.source_properties.severity = self.original_event['event']['Severity']
        # ART del(finding.source_properties.CommandLine)
        # ART del(finding.source_properties.FalconEventId)

    @property
    def event_category(self):
        # ART Uncomment the following lines to force behavior parity with the prior art. Don't.
        # ART tactic = self.event.original_event['event']['Tactic']
        # ART technique = self.event.original_event['event']['Technique']
        # ART if tactic and technique:
        #    return 'Namespace: TTPs, Category: {}, Classifier: {}'.format(tactic, technique)
        # ART return self.event.detect_description
        return self.event.detect_name

    @property
    def severity(self):
        sev = self.event.severity.upper()
        return 'LOW' if sev == 'INFORMATIONAL' else sev

    def submit_finding(self, finding):
        return self.cache.submit_finding(self.finding_id, finding, self.org_id)

    @property
    def asset_path(self):
        return self.asset.name

    @property
    def asset(self) -> Asset:
        return self.cache.asset(self.event)

    @property
    def finding_path(self):
        return SecurityCenterClient.finding_path(self.org_id, self.source_id, self.finding_id)

    @property
    @lru_cache
    def finding_id(self):
        event_id = re.sub('[^0-9a-zA-Z]+', '', self.event.event_id)
        creation_time = str(hex(int(self.event.original_event['metadata']['eventCreationTime'])))[-6:]
        return (event_id + creation_time)[-32:]

    @property
    def source_path(self):
        return SecurityCenterClient.source_path(self.org_id, self.source_id)

    @property
    def source_id(self):
        parsed = SecurityCenterClient.parse_source_path(self.source.name)
        return parsed['source']

    @property
    @lru_cache
    def source(self):
        return self.cache.source(self.org_id)

    @property
    @lru_cache
    def org_id(self):
        return self.cache.organization_parent_of(self.gcp_project_number)

    @property
    @lru_cache
    def gcp_project_number(self):
        return self.event.cloud_provider_account_id


class Runtime():
    RELEVANT_EVENT_TYPES = ['DetectionSummaryEvent']

    def __init__(self):
        log.info("GCP Backend is enabled.")
        self.cache = Cache()

    def is_relevant(self, falcon_event):
        return falcon_event.cloud_provider == 'GCP'

    def process(self, falcon_event):
        Submitter(self.cache, falcon_event).submit()


__all__ = ['Runtime']



from logging import INFO, Formatter, getLogger
from ssl import CERT_NONE, PROTOCOL_TLSv1_2
from tlssyslog.handlers import TLSSysLogHandler
from ....config import config

serverlog = getLogger('ws1')
serverlog.setLevel(INFO)

tlsHandler = TLSSysLogHandler(
    address=(config.get('workspaceone', 'syslog_host'), int(config.get('workspaceone', 'syslog_port'))),
    ssl_kwargs={'cert_reqs': CERT_NONE, 'ssl_version': PROTOCOL_TLSv1_2}
)
tlsHandler.setLevel(INFO)

formatter = Formatter('%(asctime)s %(name)s %(threadName)-10s %(levelname)-8s %(message)s', '%Y-%m-%d %H:%M:%S')
tlsHandler.setFormatter(formatter)

serverlog.addHandler(tlsHandler)



from google.cloud.resourcemanager import FoldersClient, Project, ProjectsClient
from .scc import SecurityCommandCenter


def project(project_number: int):
    client = ProjectsClient()
    return client.get_project(name="projects/" + project_number)


def folder(folder_path: str):
    client = FoldersClient()
    return client.get_folder(name=folder_path)


def project_get_parent_org(prj: Project):
    client = ProjectsClient()
    item = prj
    while True:
        parent = item.parent
        parsed = client.parse_common_organization_path(parent)
        if parsed:
            return parsed['organization']
        parsed = client.parse_common_folder_path(parent)
        if parsed:
            item = folder(parent)
        else:
            raise Exception('Could not detemine parent organization for GCP project {}, last parent determined was: {}'.format(prj.name, parent))


__all__ = ['project', 'SecurityCommandCenter']



from datetime import datetime
import traceback
import boto3
from botocore.exceptions import ClientError
from ...config import config
from ...log import log


class Submitter():
    def __init__(self, event):
        self.event = event

    def find_instance(self, instance_id, mac_address):
        # Instance IDs are unique to the region, not the account, so we have to check them all
        report_region = config.get('aws', 'region')
        ec2instance = None
        ec2_client = boto3.client("ec2")
        regions = [region["RegionName"] for region in ec2_client.describe_regions()["Regions"]]
        for region in regions:
            ec2 = boto3.resource("ec2", region_name=region)
            try:
                ec2instance = ec2.Instance(instance_id)
                found = False
                # Confirm the mac address matches
                for iface in ec2instance.network_interfaces:
                    det_mac = mac_address.lower().replace(":", "").replace("-", "")
                    ins_mac = iface.mac_address.lower().replace(":", "").replace("-", "")
                    if det_mac == ins_mac:
                        found = True
                if found:  # pylint: disable=R1723
                    return region, ec2instance
            except ClientError:
                continue
            except Exception:  # pylint: disable=W0703
                trace = traceback.format_exc()
                log.exception(str(trace))
                continue

        return report_region, ec2instance

    @staticmethod
    def send_to_securityhub(manifest):
        client = boto3.client('securityhub', region_name=config.get('aws', 'region'))
        check_response = {}
        found = False
        try:
            check_response = client.get_findings(Filters={'Id': [{'Value': manifest["Id"], 'Comparison': 'EQUALS'}]})
            for _ in check_response["Findings"]:
                found = True
        except ClientError:
            pass

        import_response = False
        if not found:
            try:
                import_response = client.batch_import_findings(Findings=[manifest])
            except ClientError as err:
                # Boto3 issue communicating with SH, throw the error in the log
                log.exception(str(err))

        return import_response

    def submit(self):
        log.info("Processing detection: %s", self.event.detect_description)
        det_region = config.get('aws', 'region')
        send = False
        try:
            if self.event.instance_id:
                det_region, instance = self.find_instance(self.event.instance_id, self.event.device_details["mac_address"])
                if instance is None:
                    log.warning("Instance %s with MAC address %s not found in regions searched. Alert not processed.",
                                self.event.instance_id, self.event.device_details["mac_address"])
                    return
                try:
                    for _ in instance.network_interfaces:
                        # Only send alerts for instances we can find
                        send = True

                except ClientError:
                    # Not our instance
                    i_id = self.event.instance_id
                    mac = self.event.device_details["mac_address"]
                    log.info("Instance %s with MAC address %s not found in regions searched. Alert not processed.", i_id, mac)
        except AttributeError:
            # Instance ID was not provided by the detection
            log.info("Instance ID not provided by detection. Alert not processed.")

        if send:
            sh_payload = self.create_payload(det_region)
            response = self.send_to_securityhub(sh_payload)
            if not response:
                log.info("Detection already submitted to Security Hub. Alert not processed.")
            else:
                if response["SuccessCount"] > 0:
                    submit_msg = f"Detection submitted to Security Hub. (Request ID: {response['ResponseMetadata']['RequestId']})"
                    log.info(submit_msg)

    def create_payload(self, instance_region):
        region = config.get('aws', 'region')
        try:
            account_id = boto3.client("sts").get_caller_identity().get('Account')
        except KeyError:
            # Failed to get endpoint_resolver the first time, try it again
            account_id = boto3.client("sts").get_caller_identity().get('Account')
        severity_product = self.event.severity_value
        severity_normalized = severity_product * 20
        payload = {
            "SchemaVersion": "2018-10-08",
            "ProductArn": "arn:aws:securityhub:{}:517716713836:product/crowdstrike/crowdstrike-falcon".format(region),
            "AwsAccountId": account_id,
            "SourceUrl": self.event.falcon_link,
            "GeneratorId": "Falcon Host",
            "CreatedAt": datetime.utcfromtimestamp(float(self.event.event_create_time) / 1000.).isoformat() + 'Z',
            "UpdatedAt": ((datetime.utcfromtimestamp(datetime.timestamp(datetime.now()))).isoformat() + 'Z'),
            "RecordState": "ACTIVE",
            "Severity": {"Product": severity_product, "Normalized": severity_normalized}
        }

        # Instance ID based detail
        try:
            payload["Id"] = f"{self.event.instance_id}{self.event.event_id}"
            payload["Title"] = "Falcon Alert. Instance: %s" % self.event.instance_id
            payload["Resources"] = [{"Type": "AwsEc2Instnace", "Id": self.event.instance_id, "Region": instance_region}]
        except AttributeError:
            payload["Id"] = f"UnknownInstanceID:{self.event.event_id}"
            payload["Title"] = "Falcon Alert"
            payload["Resources"] = [{"Type": "Other",
                                     "Id": f"UnknownInstanceId:{self.event.event_id}",
                                     "Region": region
                                     }]
        # Description
        aws_id = ""
        if self.event.cloud_provider_account_id:
            aws_id = f"| AWS Account for alerting instance: {self.event.cloud_provider_account_id}"
        payload["Description"] = f"{self.event.detect_description} {aws_id}"

        # TTPs
        try:
            payload["Types"] = ["Namespace: TTPs",
                                "Category: %s" % self.event.original_event["event"]["Tactic"],
                                "Classifier: %s" % self.event.original_event["event"]["Technique"]
                                ]
        except KeyError:
            payload.pop("Types", None)

        # Running process detail
        try:
            payload["Process"] = {}
            payload["Process"]["Name"] = self.event.original_event["event"]["FileName"]
            payload["Process"]["Path"] = self.event.original_event["event"]["FilePath"]
        except KeyError:
            payload.pop("Process", None)

        # Network detail
        try:
            payload['Network'] = self.network_payload()
        except KeyError:
            pass

        return payload

    def network_payload(self):
        net = {}
        net['Direction'] = \
            "IN" if self.event.original_event['event']['NetworkAccesses'][0]['ConnectionDirection'] == 0 else 'OUT'
        net['Protocol'] = self.event.original_event['event']['NetworkAccesses'][0]['Protocol']
        net['SourceIpV4'] = self.event.original_event['event']['NetworkAccesses'][0]['LocalAddress']
        net['SourcePort'] = self.event.original_event['event']['NetworkAccesses'][0]['LocalPort']
        net['DestinationIpV4'] = self.event.original_event['event']['NetworkAccesses'][0]['RemoteAddress']
        net['DestinationPort'] = self.event.original_event['event']['NetworkAccesses'][0]['RemotePort']
        return net


class Runtime():
    RELEVANT_EVENT_TYPES = ['DetectionSummaryEvent']

    def __init__(self):
        log.info("AWS Backend is enabled.")

    def is_relevant(self, falcon_event):
        return falcon_event.cloud_provider[:3].upper() == 'AWS'

    def process(self, falcon_event):
        Submitter(falcon_event).submit()


__all__ = ['Runtime']



import json
import re
import datetime
from ..config import config
from ..log import log


class Event(dict):
    def __init__(self, event_string, feed_id):
        event = json.loads(event_string.decode('utf-8'))
        self.feed_id = feed_id
        super().__init__(event)

    def __eq__(self, other):
        if not isinstance(other, Event):
            return False
        return super().__eq__(self, other) and self.feed_id == other.feed_id

    def irrelevant(self):
        decision = self.severity < int(config.get('events', 'severity_threshold')) \
            or self.creation_time < self.cut_off_date()
        if decision:
            log.debug("A detection event has been skipped based on severity_threshold of cut_off_date settings")
        return decision

    @property
    def event_type(self):
        return self['metadata']['eventType']

    @property
    def offset(self):
        return self['metadata']['offset']

    @property
    def uid(self):
        return str(self.feed_id) + '_' + str(self.offset)

    @property
    def severity(self):
        return self['event'].get('Severity', 5)

    @property
    def creation_time(self):
        return self.parse_cs_time(self['metadata']['eventCreationTime'])

    @property
    def sensor_id(self):
        return self['event']['SensorId']

    @property
    def computer_name(self):
        return self['event']['ComputerName']

    @classmethod
    def parse_cs_time(cls, cs_timestamp):
        return datetime.datetime.utcfromtimestamp(float(cs_timestamp) / 1000.0)

    @classmethod
    def cut_off_date(cls):
        return datetime.datetime.now() - datetime.timedelta(days=int(config.get('events', 'older_than_days_threshold')))


class Stream(dict):
    @property
    def token(self):
        return self['sessionToken']['token']

    @property
    def url(self):
        return self['dataFeedURL']

    @property
    def refresh_interval(self):
        return self['refreshActiveSessionInterval']

    @property
    def partition(self):
        match = re.match(r'.*\/sensors\/entities\/datafeed-actions/v1/([0-9a-zA-Z]+)\?',
                         self['refreshActiveSessionURL'])
        if not match or not match.group(1):
            raise Exception('Cannot parse stream partition from stream data: {}'.format(self))
        return match.group(1)

    @property
    def feed_id(self):
        if 'feed_id' not in self:
            match = re.match(r'.*\/sensors\/entities\/datafeed/v1/([0-9a-zA-Z]+)\?',
                             self['dataFeedURL'])
            if not match or not match.group(1):
                raise Exception('Cannot parse Feed ID from stream data: {}'.format(self))
            self['feed_id'] = match.group(1)
        return self['feed_id']



from setuptools import find_packages
from setuptools import setup
from glob import glob
from os.path import basename
from os.path import splitext
import fig

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="falcon-integration-gateway",
    version=fig.__version__,
    author="CRWD Solution Architects",
    author_email="integrations@crowdstrike.com",
    description="The CrowdStrike Demo Falcon Integration Gateway for GCP",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/crowdstrike/falcon-integration-gateway",
    packages=find_packages("fig"),
    package_dir={"": "fig"},
    py_modules=[splitext(basename(path))[0] for path in glob("fig/*.py")],
    include_package_data=True,
    install_requires=[
        'boto3',
        'crowdstrike-falconpy',
        'google-cloud-securitycenter',
        'google-cloud-resource-manager >= 1.0.2',
        'tls-syslog',
        'google-auth',
        'google-api-python-client',
        'py7zr'
    ],
    extras_require={
        'devel': [
            'flake8',
            'pylint',
            'pytest',
            'bandit',
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: The Unlicense (Unlicense)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import logging
from ..config import config

level = logging.getLevelName(config.get('logging', 'level'))

log = logging.getLogger('fig')
log.setLevel(level)

ch = logging.StreamHandler()
ch.setLevel(level)

formatter = logging.Formatter('%(asctime)s %(name)s %(threadName)-10s %(levelname)-8s %(message)s', '%Y-%m-%d %H:%M:%S')
ch.setFormatter(formatter)
log.addHandler(ch)



import queue
import threading


class FalconEvents(queue.Queue):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.offset = {}
        self._lock = threading.Lock()

    def last_offset(self, feed_id):
        return self.offset.get(feed_id, 0)

    def get(self, *args, **kwargs):
        event = super().get(*args, **kwargs)
        feed_id = event.feed_id
        offset = event.offset
        with self._lock:
            self.offset[feed_id] = max(offset, self.last_offset(feed_id))
        return event


falcon_events = FalconEvents()



from .serverlog import serverlog
from ...log import log
from ...config import config


class Submitter():
    def __init__(self, token, event):
        self.workspaceone_token = token
        self.event = event

    def submit(self):
        log.info("Processing detection: %s", self.event.detect_description)
        serverlog.info(self.log())

    def log(self):
        event = self.event.original_event['event']
        meta = self.event.original_event['metadata']
        msg = 'CEF:0|CrowdStrike|FalconHost|1.0|DetectionSummaryEvent|Detection Summary Event|2|'
        msg += 'Token=' + self.workspaceone_token
        msg += ' UDID=' + self.event.mdm_identifier
        if 'SensorId' in event:
            msg += ' externalId=' + event['SensorId']
        if 'ProcessId' in event:
            msg += ' cn2=' + str(event['ProcessId'])
            msg += ' cn2Label=ProcessId'
        if 'ParentProcessId' in event:
            msg += ' cn1=' + str(event['ParentProcessId'])
            msg += ' cn1Label=ParentProcessId'
        if 'UserName' in event:
            msg += ' suser=' + event['UserName']
        if 'DetectDescription' in event:
            msg += ' msg=' + event['DetectDescription']
        if 'FileName' in event:
            msg += ' fname=' + event['FileName']
        if 'FilePath' in event:
            msg += ' filePath=' + event['FilePath']
        if 'CommandLine' in event:
            msg += ' cs5=' + event['CommandLine']
            msg += ' cs5Label=CommandLine'
        if 'MD5String' in event:
            msg += ' fileHash=' + event['MD5String']
        if 'MachineDomain' in event:
            msg += ' sntdom=' + event['MachineDomain']
        if 'FalconHostLink' in event:
            msg += ' cs6=' + event['FalconHostLink']
            msg += ' cs6Label=FalconHostLink'
        if 'offset' in meta:
            msg += ' cn3=' + str(meta['offset'])
            msg += ' cn3Label=Offset'
        if 'eventCreationTime' in meta:
            msg += ' rt=' + str(meta['eventCreationTime'])
        if 'LocalIP' in event:
            msg += ' src=' + event['LocalIP']
        if 'MACAddress' in event:
            msg += ' smac=' + event['MACAddress']
        if 'Tactic' in event:
            msg += ' cat=' + event['Tactic']
        if 'Technique' in event:
            msg += ' act=' + event['Technique']
        if 'Objective' in event:
            msg += ' reason=' + event['Objective']
        if 'PatternDispositionValue' in event:
            msg += ' outcome=' + str(event['PatternDispositionValue'])
        if 'PatternDispositionDescription' in event:
            msg += ' CSMTRPatternDisposition=' + event['PatternDispositionDescription']
        return msg


class Runtime():
    RELEVANT_EVENT_TYPES = ['DetectionSummaryEvent']

    def __init__(self):
        log.info("Workspace One backend is enabled.")
        self.workspaceone_token = config.get('workspaceone', 'token')

    def is_relevant(self, falcon_event):
        return falcon_event.mdm_identifier is not None

    def process(self, falcon_event):
        Submitter(self.workspaceone_token, falcon_event).submit()


__all__ = ['Runtime']



from datetime import datetime
import ipaddress
from urllib.parse import quote
from json import dumps
from google.oauth2 import service_account
from google.auth.transport.requests import AuthorizedSession
from ...log import log
from ...config import config

SCOPES = ['https://www.googleapis.com/auth/malachite-ingestion']


def parse_url(url):
    try:
        cid = url.split("_")[-1]
        segments = url.split("/")
        relevant_url = ""
        for i in range(3, len(segments)):
            relevant_url += "/" + segments[i]
        parsed_relevant_url = quote(relevant_url, safe='')
        final_url = segments[0] + "/"
        final_url += segments[1] + "/"
        final_url += segments[2] + "/api2/link?"
        final_url += cid + "&url=" + parsed_relevant_url
        return final_url.split("_")[0]
    except IndexError as error:
        log.error("Failed to parse FalconHostLink: %s", error)
        return None


# Computes the size of a string in bytes
def utf8len(my_string):
    return len(my_string.encode('utf-8'))


class Submitter():
    def __init__(self):
        self.region = config.get('chronicle', 'region')
        self.customer_id = config.get('chronicle', 'customer_id')
        self.http_client = self.build_http_client()
        self.event = None

    def submit(self, event):
        self.event = event
        log.info("Processing detection: %s", self.event.original_event['event']['DetectId'])
        self.post_to_chronicle(self.udm(), self.region, self.customer_id)

    # Maps detection events into UDM
    def udm(self):
        event = self.event.original_event['event']
        meta = self.event.original_event['metadata']
        timestamp_components = str(datetime.fromtimestamp(meta["eventCreationTime"] / 1000)).split()
        new_url = parse_url(event["FalconHostLink"])
        udm_result = {
            "metadata": {
                "event_timestamp": timestamp_components[0] + "T" + timestamp_components[1] + "+00:00",
                "event_type": "PROCESS_UNCATEGORIZED",
                "description": event["DetectDescription"],
                "product_event_type": meta["eventType"],
                "product_log_id": event["DetectId"],
                "product_name": "Falcon",
                "vendor_name": "CrowdStrike"
            },
            "principal": {
                "asset_id": "CrowdStrike.Falcon:" + event["SensorId"],
                "hostname": self.event.original_event.computer_name,
                "user": {
                    "userid": event["UserName"]
                },
                "ip": self.check_ip(event["LocalIP"])
            },
            "target": {
                "process": {
                    "command_line": event["CommandLine"],
                    "file": {
                        "full_path": event["FilePath"] + "\\" + event["FileName"],
                        "md5": event["MD5String"],
                        "sha1": event["SHA1String"],
                        "sha256": event["SHA256String"]
                    },
                    "pid": str(event["ProcessId"]),
                    "parent_process": {
                        "command_line": event["ParentCommandLine"],
                        "pid": str(event["ParentProcessId"])
                    }
                }
            },
            "security_result": {
                "action_details": event["PatternDispositionDescription"],
                "severity_details": event["SeverityName"],
                "url_back_to_product": new_url
            }
        }
        return udm_result

    @staticmethod
    def check_ip(data):
        try:
            ipaddress.ip_address(data)
            return data
        except ValueError:
            return "127.0.0.1"

    # Posts detection to Chronicle
    def post_to_chronicle(self, detection, region, customer_id):
        # Only Chronicle's US region doesn't have an API prefix
        prefix = region + '-'
        if region == 'us':
            prefix = ''
        # Set up the url and header
        url = "https://" + prefix + "malachiteingestion-pa.googleapis.com/v2/udmevents:batchCreate"
        headers = {'Content-Type': 'application/json'}
        # Format the events field
        payload = {"customer_id": customer_id, "events": [detection]}
        # Post to Chronicle
        response = self.http_client.post(url, data=dumps(payload), headers=headers)
        # Log any errors
        if response.status_code < 200 or response.status_code > 299:
            log.error("Error posting %s to Chronicle: %s", detection['metadata']['product_log_id'], response.text)
            log.error("Faulty detection: %s", dumps(payload, indent=4, sort_keys=True))
        else:
            log.info("Successfully posted %s to Chronicle:\t Byte count: %s", detection['metadata']['product_log_id'], utf8len(dumps(payload)))

    @staticmethod
    def build_http_client():
        service_account_file = config.get('chronicle', 'service_account')
        # get google token
        credentials = service_account.Credentials.from_service_account_file(service_account_file, scopes=SCOPES)
        # Build an HTTP client to make authorized OAuth requests.
        return AuthorizedSession(credentials)


class Runtime():
    RELEVANT_EVENT_TYPES = ['DetectionSummaryEvent']

    def __init__(self):
        log.info("Chronicle backend is enabled.")
        self.submitter = Submitter()

    def is_relevant(self, falcon_event):
        return True

    def process(self, falcon_event):
        self.submitter.submit(falcon_event)


__all__ = ['Runtime']



import threading
from google.api_core.exceptions import AlreadyExists
from google.cloud import securitycenter
from google.cloud.securitycenter import CreateFindingRequest, Finding, Source
from google.protobuf.field_mask_pb2 import FieldMask  # pylint: disable=no-name-in-module
from ....log import log


_sources_lock = threading.Lock()


class SecurityCommandCenter():
    FIG_SOURCE_NAME = 'CrowdStrike Falcon'

    def __init__(self):
        self.client = securitycenter.SecurityCenterClient()

    def get_asset(self, project_number, asset_id):
        asset_iterator = self.client.list_assets(request={
            "parent": self.client.common_project_path(project_number),
            'filter': 'resource_properties.id="{}"'.format(asset_id)
        })
        return list(asset_iterator)

    def list_instances(self, project_number):
        asset_iterator = self.client.list_assets(request={
            "parent": self.client.common_project_path(project_number),
            'filter': 'security_center_properties.resource_type=' + '"google.compute.Instance"',
        })
        return list(asset_iterator)

    def get_fig_source(self, org_id):
        for _, source in enumerate(self.client.list_sources(request={"parent": self._org_name(org_id)})):
            if source.display_name == self.FIG_SOURCE_NAME:
                return source
        return None

    def get_or_create_fig_source(self, org_id):
        source = self.get_fig_source(org_id)
        if source is not None:
            return source

        with _sources_lock:
            source = self.get_fig_source(org_id)
            if source is not None:
                return source
            return self.create_fig_source(org_id)

    def create_fig_source(self, org_id):
        log.info("Creating new Finding Source in GCP Security Command Center (org_id=%s)", org_id)
        return self.client.create_source(
            request={
                "parent": self._org_name(org_id),
                "source": {
                    "display_name": self.FIG_SOURCE_NAME,
                    "description": "CrowdStrike Falcon findings forwarded by falcon-integration-gateway",
                },
            }
        )

    def get_or_create_finding(self, finding_id, finding: Finding, source: Source):
        existing = self.get_finding(finding, source)
        if len(existing) == 0:
            try:
                return [self.create_finding(finding_id, finding, source)]
            except AlreadyExists:
                existing = self.get_finding(finding, source)
                if len(existing) == 0:
                    raise

        log.debug("Finding %s already exists in GCP SCC", finding_id)
        return existing[0].finding

    def get_finding(self, finding: Finding, source: Source):
        return list(self.client.list_findings(
            request={
                'parent': source.name,
                'filter': 'name="{}"'.format(finding.name),
            }))

    def create_finding(self, finding_id, finding: Finding, source: Source):
        log.info('Submitting new finding to GCP Security Command Center (finding_id=%s)', finding_id)
        created_finding = self.client.create_finding(
            request=CreateFindingRequest(
                parent=source.name,
                finding_id=finding_id,
                finding=finding,
            )
        )
        return created_finding

    def update_finding(self, finding):
        field_mask = FieldMask(
            paths=['event_time']
        )
        updated_finding = self.client.update_finding(
            request={
                "finding": finding,
                "update_mask": field_mask,
            }
        )
        return updated_finding

    @classmethod
    def _org_name(cls, org_id):
        return 'organizations/{org_id}'.format(org_id=org_id)
